''******************************
''*  Altair CPU Simulator v2.3 *
''*  (C) 2009 Michael Rychlik  *
''******************************

' v1.0 - 01 Oct 2007 - original version
' v1.1 - 09 Dec 2007 - Moved dispatch table into HUB RAM to save COG space (and increase speed !!)
' v1.2 - 09 Mar 2008 - No change, just bump version 
' v1.3 - 20 Mar 2008 - Change do_xchg and do_xthl to use the xor swap trick with no temp storage.
'                    - memory read write overhaul , checks for RAM/ROM address limits as we don't
'                      want broken 8085 programs being able to trample over HUB RAM
'                    - Removed those flag bits from the dispatch table. Now immediate bytes words
'                      are read by in the opcode handlers that need them not in the main loop.
'                    - Removed defaulting mem_ptr to hl in main instruction loop, instructions that need it
'                      handle it themselves.
'                    - Changed push/pop_word to use memory_read/write functions.
' v1.4 - 21 Mar 2008 - Added reg_base parameter to stat method to support break points and single stepping
'                      of the emulation.                                                           
'                    - Added a heart beat display LED
'                    - Added code in main loop to single step and break cpu execution.
'                    - Bug fix DAD op code handler.
'                    - Bug fix CPI op code handler.
'                    - Made dump_regs subroutine work.
' v2.1 - 21 Apr 2008 - Too many chages to describe in detail, an almost total rewrite.
'                    - Functionaly the only new thing is implementation of the DAA instruction
'                    - The ALU is much more streamlined, easier to understand and faster
'                    - The conditional jump logic has had the same treatment
'                    - Most other instruction handlers have been streamlined somewhat
'                    - A Large Memory Model (LMM) virtual machine has been introduced to handle some instructions
'                    - Commented out break point, single step functionality as there no space.
'                    - Speed is now up from 260KIPS to 385KIPS running EX8080.COM (80MHz prop), a 48% improvemnet
'                    - over 2.0
' v2.2 - 14 Jan 2009 - Added the RST instructions. These are required to get MITS BASICS to run.
'                    - Removed the interrupt enable flag and made EI/DI into no ops to make room for RST instructions
' v2.3 - 16 Jan 2009 - Bumped version.
'                               
' 
CON                                
  'Program status word bits
  sign_bit   = %10000000
  zero_bit   = %01000000
  aux_bit    = %00010000
  parity_bit = %00000100
  carry_bit  = %00000001

  'Some Propeller PASM instruction fields
  propADD  = %100000_111
  propADDX = %110010_111
  propSUB  = %100001_111
  propSUBX = %110011_111
  propAND  = %011000_111
  propXOR  = %011011_111
  propOR   = %011010_111


VAR

  long  cog
  long  start_p
  long  end_p
  long  temp

  
PUB start(cpupar) : okay

'' Start Altair CPU simulator - starts a cog
'' returns false if no cog available
''
''   cpupar is a pointer to CPU parameter list

  reg_base := long[cpupar]
  ram_base := long[cpupar + 4]
  rom_base := long[cpupar + 8]
  ios_op   := long[cpupar + 12]
  ios_port := ios_op + 1
  ios_data := ios_op + 2

  'Reverse all the LMM Virtual Machine program instructions
  start_p := @lmm_begin
  end_p   := @lmm_end
  repeat while start_p < end_p
    temp := long[start_p]
    long[start_p] := long[end_p]
    long[end_p] := temp
    start_p += 4
    end_p -= 4

  object_base := @@0                            'Set up the base address of this object        
    
  stop
  okay := cog := cognew(@enter, cpupar) + 1    'Start emulator in a new COG

                                                        
PUB stop

'' Stop Altair CPU simulator - frees a cog

  if cog
    cogstop(cog~ - 1)


DAT
                        org     0

'Execution starts here
enter                                                      
b_reg                   mov     b_reg, #0                  'These instructions become the 8080 registers 
c_reg                   mov     c_reg, #0                  'They self initialize.
d_reg                   mov     d_reg, #0                  'Regs are put at location zero so as enable
e_reg                   mov     e_reg, #0                  'fast lookup.
h_reg                   mov     h_reg, #0
l_reg                   mov     l_reg, #0 
m_reg                   mov     m_reg, #0
a_reg                   mov     a_reg, #0
psw_reg                 mov     psw_reg, #0
pc                      mov     pc, #0
sp                      mov     sp, #0            

                        mov     dira, pins                 'Set Pins to output
                        add     dispatch_tab_addr, object_base 'Fixup this object offset to HUB real address

'Instruction fetch loop starts here
fetch                                                      

'NOP - No Operation 
do_nop

'DI - Disable Interrupts
do_di                                                      'DI - Disable Interrupts

'EI - Enable Interrupts
do_ei                                                      'EI - Enable Interrupts

do_undocumented         call    #instruction_fetch         'Get the next instruction into 'm_reg'
                        add     outa, #$100                'Display instruction counter on LEDs for speed measurement

{{
                        mov     scratch,pc                 'Check for break point
                        sub     scratch, #1
                        cmp     scratch, break_point wz 
              if_nz     jmp     #:no_trig
                        add     single_step, #1
:no_trig                 cmp     single_step, #1 wz
               if_z     call    #break
:no_break
}}
                        mov     scratch, m_reg             'Dispatch table look up
                        shl     scratch, #1                'Mul by 2 to get a WORD offset into the table
                        add     scratch, dispatch_tab_addr 'Index into table
                        rdword  scratch, scratch           'Read the instruction handlers address
                        cmp     scratch, #lmm_begin wc     'Is this a "native" or "LMM" address ?             
                if_c    jmp     scratch                    'Jump to "native" instruction handler
                        add     scratch, object_base       'Add this objects base address
                        mov     lmm_pc, scratch            'Set LMM's program counter to LMM handler
                        jmp     #lmm_start                 'Start the LMM virtual machine

'This is a reversed execution order LMM virtual machine. Courtesy of Phil Pilgrim (phipi.com)
lmm_ins_0               nop                                
                        rdlong  :lmm_ins_1, lmm_pc
                        sub     lmm_pc, #7
:lmm_ins_1              nop
lmm_start               rdlong  lmm_ins_0, lmm_pc          'Which, of course, starts near the bottom:) 
                        djnz    lmm_pc, #lmm_ins_0

'STC - Set Carry
do_stc                  or      psw_reg, #carry_bit        'STC - Set Carry
                        jmp     #fetch

'CMC - Complement Carry
do_cmc                  xor     psw_reg, #carry_bit        'CMC - Complement Carry
                        jmp     #fetch
                  
'HALT - Halt
do_halt                 jmp     #do_halt                   'HALT


'IN - Input
do_in

'OUT - Output
do_out
do_io                   mov     opc, m_reg                 'Extract IN or OUT operation from instruction byte
                        shr     opc, #3
                        and     opc, #7
                        call    #instruction_fetch         'Get i/o port number from next instruction byte
                        wrbyte  m_reg, ios_port            'Set i/o subsytems port number
                        wrbyte  a_reg, ios_data            'Set i/o subsystems data from accumulator
                        wrbyte  opc, ios_op                'Set i/o subsystems operation from the opcode field
:io_not_rdy             rdbyte  scratch, ios_op wz         'Loop until i/o susbsystem clears the operation
              if_nz     jmp     #:io_not_rdy
                        rdbyte  a_reg, ios_data            'Read back the data into acc. incase this was an IN.
                        jmp     #fetch

'CMA - Complement Accumulator
do_cma                  xor     a_reg, #$FF                
                        jmp     #fetch

'STAX B - Store accumulator at (BC)
do_stax_b               mov     address, b_reg             'Get "address" from B and C             
                        shl     address, #8
                        add     address, c_reg
                        jmp     #stax_tail                 'Go somewhere that will store acc.
           
'STAX D - Store accumulator at (DE)
do_stax_d               mov     address, d_reg             'Get address from D and E
                        shl     address, #8
                        add     address, e_reg
stax_tail               mov     m_reg, a_reg
                        call    #write_memory              'Write byte m_reg to "address"
                        jmp     #fetch

'LDAX B - Read accumulator from (BC)
do_ldax_b               mov     address, b_reg             'Get address from B and C
                        shl     address, #8
                        add     address, c_reg
                        jmp     #ldax_tail                 'Go somewhere that will read acc.

'LDAX D - Read accumulator from (DE)
do_ldax_d               mov     address, d_reg             'Get address from D and E
                        shl     address, #8
                        add     address, e_reg
ldax_tail               call    #read_memory               'Read byte from "address" into m_reg
                        mov     a_reg, m_reg               'A load into acc.
                        jmp     #fetch

'A mini ALU just for increment and decrement
alu_inr_dcr                                                
alu_inr_dcr_load        mov     alu_reg, 0-0               'Load the alu
                        mov     aux,alu_reg                'Keep a copy for aux carry detection later
alu_inr_dcr_op          add     alu_reg, #1                'Add or subtract one (self modifying)
alu_tail                xor     aux,alu_reg                'Determine the auxillary carry flag
                        test    aux, #aux_bit wz
                        muxnz   psw_reg, #aux_bit
                        and     alu_reg, #$FF wz, wc       'Set zero, sign and parity flags from alu content
                        muxz    psw_reg, #zero_bit         'Set 8080 zero flag from props zero
                        muxnc   psw_reg, #parity_bit       'That "and" above gves parity in props carry flag
                        test    alu_reg, #128 wz           'Top bit is sign
                        muxnz   psw_reg, #sign_bit
alu_inr_dcr_unload      mov     0-0, alu_reg               'Write alu result out
alu_inr_dcr_ret         ret

'Arithmentic Logic Unit
alu                     shr     opc, #3                    'Extract the alu operator field
                        and     opc, #7
                        add     opc, #opTransTable         'Look up corresponding Prop instruction 
                        movs    :xlat, opc
                        mov     alu_reg, a_reg             'Load the alu
:xlat                   movi    alu_operator, opTransTable 'Configure the our ALU to do that op.
                        and     alu_reg, #$FF
alu_load_operand        mov     operand, 0-0               'Load the operand
                        and     operand, #$FF
                        test    psw_reg, #carry_bit wc     'Set prop's carry from 8080 carry flag
                        mov     aux,alu_reg                'Take a copy of alu before operation for AUX carry determination
alu_operator            add     alu_reg, operand           'Do the operation (may not be add !)
                        test    alu_reg, #$100 wz          'Determine the carry flag
                        muxnz   psw_reg, #carry_bit        'N.B. This will always clear carry for logical opperations

                        max     logic_op, #0 wz            'For logical operations don't set AUX carry here
              if_nz     jmp     #:zsp
                        xor     aux,operand
                        xor     aux,alu_reg                'Determine the auxillary carry flag
                        test    aux, #aux_bit wc
              if_z      muxc    psw_reg, #aux_bit

:zsp                    and     alu_reg, #$FF wz, wc       'Set zero, sign and parity flags from alu content
                        muxz    psw_reg, #zero_bit         'Set 8080 zero flag from props zero
                        muxnc   psw_reg, #parity_bit       'That "and" above gves parity in props carry flag
                        test    alu_reg, #128 wz           'Top bit is sign
                        muxnz   psw_reg, #sign_bit
alu_unload              mov     0-0, alu_reg               'Write alu result out
alu_ret                 ret

'INR [B|C|D|E|H|L|A] - increment register
do_inr_b                                                   
do_inr_c                
do_inr_d                
do_inr_e                
do_inr_h                
do_inr_l                
do_inr_a                movi    alu_inr_dcr_op, #propADD   'Modify alu_inr_dcr to do propeller ADD instruction

do_inr_dcr              shr     m_reg, #3                  'Extract register field from instruction                 
                        and     m_reg, #7
                        movs    alu_inr_dcr_load, m_reg    'Modify alu_inr_dcr to load that reg.
                        movd    alu_inr_dcr_unload, m_reg  'Modify alu_inr_dcr to unload to that reg.
                        call    #alu_inr_dcr               'Do the inrement/decrement
                        jmp     #fetch

'DCR [B|C|D|E|H|L|A] - decrement register
do_dcr_b                                                   
do_dcr_c                
do_dcr_d                
do_dcr_e                
do_dcr_h                
do_dcr_l                
do_dcr_a                movi    alu_inr_dcr_op, #propSUB   'Modify alu_inr_dcr to do propeller SUB instruction
                        jmp     #do_inr_dcr                    'Continue as increment reg

'INR M - increment memory (HL)
do_inr_m
                        movi    alu_inr_dcr_op, #propADD   'Modify alu_inr_dcr to do propeller ADD instruction
do_inr_dcr_m            call    #read_memory_hl            'Read byte from (HL) into m_reg (also sets "address")
                        movs    alu_inr_dcr_load, #m_reg   'Modify alu_inr_dcr to load that data
                        movd    alu_inr_dcr_unload, #m_reg 'Modify alu_inr_dcr to unload to that data
                        call    #alu_inr_dcr               'Do the inrement/decrement
                        call    #write_memory              'Write m_reg out to "address"      
                        jmp     #fetch

'INR M - Increment Memory (HL)
do_dcr_m                movi    alu_inr_dcr_op, #propSUB   'Modify alu_inr_dcr to do propeller SUB instruction
                        jmp     #do_inr_dcr_m              'Continue as increment memory

'RAL - Rotate accumulator Left Through Carry
do_ral                  shl     a_reg, #1                  'Shift accumulator left by one
                        test    psw_reg, #carry_bit wz     'Set acc. LSB from carry flag
                        muxnz   a_reg, #1                        
                        test    a_reg, #%1_0000_0000 wz    'Set carry flag from 8 bit overflow
                        muxnz   psw_reg, #carry_bit
                        and     a_reg, #$FF                'Keep in range
                        jmp     #fetch
             
'RAR - Rotate Accumulator Right Through Carry
do_rar                  shr     a_reg, #1 wc               'Shift accumulator right by one
                        test    psw_reg, #carry_bit wz     'Set acc. MSB from carry flag
                        muxnz   a_reg, #%10000000
                        muxc    psw_reg, #carry_bit        'Set carry flag from props carry out
                        and     a_reg, #$FF                'Keep in range
                        jmp     #fetch

'RLC - Rotate Accumulator Left
do_rlc                  shl     a_reg, #1                  'Shift accumulator left by one
                        test    a_reg, #%1_0000_0000 wz    'Move any 8 bit carry out bit into acc LSB
                        muxnz   a_reg, #1
                        muxnz   psw_reg, #carry_bit        'Set carry flag from 8 bit overflow
                        and     a_reg, #$FF                'Keep in range
                        jmp     #fetch

'RRC - Rotate Accumulator Right
do_rrc                  shr     a_reg, #1 wc               'Shift accumulator right by one
                        muxc    a_reg, #%10000000          'Move props carry to acc. MSB
                        muxc    psw_reg, #carry_bit        'Set carry flag from props carry
                        and     a_reg, #$FF                'Keep in range
                        jmp     #fetch

'MOV [B|C|D|E|H|L|A], [B|C|D|E|H|L|A] - Move Register to Register
do_mov_b_b  
do_mov_b_c  
do_mov_b_d  
do_mov_b_e
do_mov_b_h
do_mov_b_l  
do_mov_b_a
do_mov_c_b  
do_mov_c_c  
do_mov_c_d 
do_mov_c_e  
do_mov_c_h  
do_mov_c_l  
do_mov_c_a
do_mov_d_b
do_mov_d_c  
do_mov_d_d  
do_mov_d_e  
do_mov_d_h  
do_mov_d_l  
do_mov_d_a
do_mov_e_b  
do_mov_e_c  
do_mov_e_d  
do_mov_e_e  
do_mov_e_h  
do_mov_e_l  
do_mov_e_a
do_mov_h_b  
do_mov_h_c  
do_mov_h_d  
do_mov_h_e  
do_mov_h_h  
do_mov_h_l  
do_mov_h_a
do_mov_l_b  
do_mov_l_c  
do_mov_l_d  
do_mov_l_e  
do_mov_l_h  
do_mov_l_l  
do_mov_l_a
do_mov_a_b  
do_mov_a_c  
do_mov_a_d  
do_mov_a_e  
do_mov_a_h  
do_mov_a_l  
do_mov_a_a              ror     m_reg, #3                  'Extract the destination register field from the opcode
                        andn    m_reg, #%11111000
                        movd    :reg_2_reg, m_reg          'Mofify dest. of reg_2_reg to be that register
                        rol     m_reg, #3                  'Extract the source register field from the opcode
                        and     m_reg, #7
                        movs    :reg_2_reg, m_reg          'Mofify source of reg_2_reg to be that register
                        nop
:reg_2_reg              mov     0-0,0-0                    'Make the move
                        jmp     #fetch

'MOV [B|C|D|E|H|L|A], M - Move memory to register
do_mov_b_m
do_mov_c_m
do_mov_d_m
do_mov_e_m
do_mov_h_m
do_mov_l_m                                                 
do_mov_a_m              shr     m_reg, #3                  'Extract the destination register field from the opcode
                        and     m_reg, #7
                        movd    :m_reg_2_reg, m_reg        'Mofify dest. of m_reg_2_reg to be that register
                        call    #read_memory_hl            'Read byte from (HL) into m_reg
:m_reg_2_reg            mov     0-0, m_reg                 'Write m_reg to the selected register
                        jmp     #fetch

'MOV M,[B|C|D|E|H|L|A] - Move Register To Memory 
do_mov_m_b  
do_mov_m_c  
do_mov_m_d  
do_mov_m_e
do_mov_m_h  
do_mov_m_l  
do_mov_m_a              and     m_reg, #7                  'Extract the source register field from the opcode
                        movs    :reg_2_m_reg, m_reg        'Mofify source of reg2m_reg to be that reg 
                        nop
:reg_2_m_reg            mov     m_reg, 0-0
                        call    #write_memory_hl           'Write byte m_reg into memory (HL)
                        jmp     #fetch

'MVI [B|C|D|E|H|L|A] - Move immediate to register
do_mvi_b
do_mvi_c
do_mvi_d
do_mvi_e
do_mvi_h
do_mvi_l                                                   'Extract the register field from the opcode
do_mvi_a                shr     m_reg, #3
                        and     m_reg, #7
                        movd    :m_reg_2_reg, m_reg        'Modify :m_reg_2_reg to save to that register
                        call    #instruction_fetch         'Get the instructions immediate byte into m_reg 
:m_reg_2_reg            mov     0-0, m_reg                 'Put m_reg to selected register
                        jmp     #fetch

'MVI M - Move Immediate to Memory
do_mvi_m                call    #instruction_fetch         'Get the instructions immediate byte into m_reg  
                        call    #write_memory_hl           'Write m_reg to (HL)
                        jmp     #fetch

'ANA [B|C|D|E|H|L|A] - Logical AND Register with Accumulator
do_ana_b
do_ana_c
do_ana_d
do_ana_e
do_ana_h
do_ana_l
do_ana_a

'XRA [B|C|D|E|H|L|A] - Logical Exclusive OR Register with Accumulator
do_xra_b
do_xra_c
do_xra_d
do_xra_e
do_xra_h
do_xra_l
do_xra_a                

'ORA [B|C|D|E|H|L|A] - Logical OR Register With Accumulator
do_ora_b
do_ora_c
do_ora_d
do_ora_e
do_ora_h
do_ora_l
do_ora_a                mov     logic_op, #1

'ADD [B|C|D|E|H|L|A] - Add Register To Accumulator
do_add_b                
do_add_c
do_add_d      
do_add_e  
do_add_h  
do_add_l
do_add_a                

'ADC [B|C|D|E|H|L|A] - Add Register To Accumulator With Carry
do_adc_b
do_adc_c
do_adc_d
do_adc_e
do_adc_h
do_adc_l
do_adc_a                

'SUB [B|C|D|E|H|L|A] - Subtract Register From Accumulator
do_sub_b
do_sub_c
do_sub_d
do_sub_e
do_sub_h
do_sub_l
do_sub_a                

'SBB [B|C|D|E|H|L|A] - Subtract Register From Accumulator With Borrow
do_sbb_b
do_sbb_c
do_sbb_d
do_sbb_e
do_sbb_h
do_sbb_l
do_sbb_a
                        movd    alu_unload, #a_reg
reg_acc_tail            mov     opc, m_reg                 'Extract register field from opcode
                        and     opc, #7
                        movs    alu_load_operand, opc      'Set up ALU's operand load register index
                        mov     opc, m_reg                 'Set up opcode for ALU
                        call    #alu
                        jmp     #fetch             

'CMP [B|C|D|E|H|L|A] - Compare Register With Accumulator
do_cmp_b
do_cmp_c
do_cmp_d
do_cmp_e
do_cmp_h
do_cmp_l
do_cmp_a 
                        movd    alu_unload, #scratch       'Discard the result, then ALU gets the flags right.
                        jmp     #reg_acc_tail

'ANA M - Logical AND Acumulator With Memory
do_ana_m

'XRA M - Logical Exclusive OR Accumulator With Memory
do_xra_m

'ORA M - Logical OR Accumulator With Memory
do_ora_m
                        mov     logic_op, #1

'ADD M - Add Memory to Accumulator
do_add_m

'ADD M - Add Memory to Accumulator With Carry
do_adc_m

'SUB M - Subtract Memory from Accumulator
do_sub_m

'SBB M - Subtract Memory from Accumulator With Borrow
do_sbb_m
                        movd    alu_unload, #a_reg
mem_acc_tail            mov     opc,m_reg                  'Save opcode for ALU
                        call    #read_memory_hl            'Read byte from (HL) into m_reg, also sets "address"
                        movs    alu_load_operand, #m_reg   'Set up ALU'soperand load register index
                        call    #alu
                        call    #write_memory              'FIX ME, CMP should not write back
                        jmp     #fetch

'CMP M - Compare Memory with Accumulator
do_cmp_m                movd    alu_unload, #scratch       'Discard the result, then ALU gets the flags right.
                        jmp     #mem_acc_tail


'ANI - Logical AND Acumulator With Immediate
do_ani                  

'XRI - Logical Exclusive OR Acumulator With Immediate
do_xri                  

'ORI - Logical OR Acumulator With Immediate
do_ori
                        mov     logic_op, #1

'ADI - Add Immediate To Accumulator
do_adi                  

'ADI - Add Immediate To Accumulator With Carry
do_aci                  

'SUI - Subtract Immediate From Accumulator
do_sui                  

'SBI - Subtract Immediate From Accumulator With Borrow
do_sbi                  
                        movd    alu_unload, #a_reg
imm_acc_tail            mov     opc, m_reg                 'Set opcode for ALU
                        call    #instruction_fetch         'Get the instructions imediate byte into m_reg
                        movs    alu_load_operand, #m_reg   'Set up ALU's operand load register index
                        call    #alu
                        jmp     #fetch

'CPI - Compare Immediate With Accumulator
do_cpi                  movd    alu_unload, #scratch
                        jmp     #imm_acc_tail                 

'LHLD - Load H And L Direct
do_lhld                 call    #word_fetch                'Get instructions immediate word into "address"
                        call    #read_memory               'Read byte from "address" into m_reg
                        mov     l_reg, m_reg        
                        add     address, #1                'Address the next byte
                        call    #read_memory               'Read byte from "address" into m_reg
                        mov     h_reg, m_reg
                        jmp     #fetch

'SHLD - Store H And L Direct
do_shld                 call    #word_fetch                'Get instructions immediate word into "address"
                        mov     m_reg, l_reg               'Set m_reg param of write_memory to "l" reg
                        call    #write_memory              'Write out the "l" reg to "address"
                        add     address, #1            
                        mov     m_reg, h_reg               'Set m_reg param of write_memory to "h" reg
                        call    #write_memory              'Write out the "h" reg to "address"
                        jmp     #fetch                 

'STA - Store Accumulator Direct
do_sta                  call    #word_fetch                'Get instructions immediate word into "address"
                        mov     m_reg, a_reg               'Set m_reg param of write_memory to "a" reg
                        call    #write_memory              'Write out the "a" reg to "address"
                        jmp     #fetch

'LDA - Load Accumulator Direct
do_lda                  call    #word_fetch                'Get instructions immediate word into "address"
                        call    #read_memory               'Read a byte from "address" into m_reg
                        mov     a_reg, m_reg               'Get the read byte in "a" reg
                        jmp     #fetch

'INX SP - Increment Stack Pointer                        
do_inx_sp               add     sp, #1
                        jmp     #fetch

'DCX SP - Decrement Stack Pointer                        
do_dcx_sp               sub     sp, #1
                        jmp     #fetch

'SPHL - Load Stack Pointer From H and L                        
do_sphl                 mov     sp, h_reg
                        shl     sp, #8
                        add     sp, l_reg
                        jmp     #fetch

'PCHL - Load Program Counter From H and L                        
do_pchl                 mov     pc, h_reg
                        shl     pc, #8
                        add     pc, l_reg
                        jmp     #fetch

'RNZ - Return If Not Zero
do_rnz

'CNZ - Call If Not Zero
do_cnz

'JNZ - Jump If Not Zero
do_jnz                  test    psw_reg, #zero_bit wz
                        jmp     #doitnot

'RZ - Return If Zero
do_rz

'JZ - Jump If Zero
do_jz

'CZ - Call If Zero
do_cz                   test    psw_reg, #zero_bit wz
                        jmp     #doit

'RNC - Return If Not Carry
do_rnc

'CNC - Call If Not Carry
do_cnc

'JNC - Jump If Not Carry
do_jnc                  test    psw_reg, #carry_bit wz
                        jmp     #doitnot

'RC - Return If Carry
do_rc

'CC - Call If Carry
do_cc

'JC - Jump If Carry
do_jc                   test    psw_reg, #carry_bit wz
                        jmp     #doit

'RPO - Return If Parity Odd
do_rpo

'CPO - Call If Parity Odd
do_cpo

'JPO - Call If Parity Odd
do_jpo                  test    psw_reg, #parity_bit wz
                        jmp     #doitnot

'RPE - Return If Parity Even
do_rpe

'CPE - Call If Parity Even
do_cpe

'JPE - Jump If Parity Even
do_jpe                  test    psw_reg, #parity_bit wz
                        jmp     #doit

'RM - Return If Minus
do_rm

'CM - Call If Minus
do_cm

'JM - Jump If Minus
do_jm                   test    psw_reg, #sign_bit wz
                        jmp     #doit

'RP - Return If Plus
do_rp

'CP - Call If Plus
do_cp

'JP - Jump If Plus
do_jp                   test    psw_reg, #sign_bit wz

doitnot 'Do ret, call or jmp if IF_Z
                        muxz    do_flag, #1
                        jmp     #cont

doit  'Do ret, call or jmp IF_NZ
                        muxnz   do_flag, #1
cont                    shr     m_reg, #1                  'Extract jmp, call, ret field of instruction
                        and     m_reg, #3
                        cmp     m_reg ,#1 wc, wz                                
              if_z      jmp     #jmp_conditional           '1 is conditional jump
              if_c      jmp     #ret_conditional           '0 is conditional return
                                                           '2 is conditional call, fall through.
call_conditional        call    #word_fetch                'Get instructions immediate word into "address"
                        test    do_flag, #1 wz
              if_z      jmp     #fetch

call_conditional_tail   mov     scratch, address

call_scratch            sub     sp, #1                     'Decrement stack pointer first
                        and     sp, word_mask              'Ensure still 16 bits (roll over)
                        mov     address,sp                 'Set "address" parameter of write_memory to stack pointer
                        mov     m_reg, pc                  'Set m_reg parameter of write_memory to "scratch"
                        ror     m_reg, #8
                        call    #write_memory              'Write the high byte to the stack
                        sub     sp, #1                     'Decrement stack pointer again
                        and     sp, word_mask              'Ensure still 16 bits (roll over)
                        mov     address,sp                 'Set "address" parameter of write_memory to stack pointer
                        rol     m_reg, #8
                        call    #write_memory              'Write the low byte to the stack
                        mov     pc, scratch                'Make the jump
                        jmp     #fetch

ret_conditional         test    do_flag, #1 wz
              if_z      jmp     #fetch                     'Fall through to ret if condition met

'RET - Return
do_ret                  mov     address, sp                'Set "address" param or read_word to sp    
                        call    #read_memory               'Read a byte from the stack into m_reg
                        mov     pc, m_reg
                        ror     pc, #8
                        add     sp, #1                     'Bump up the stack pointer
                        and     sp, word_mask              'Ensure still 16 bits (roll over)
                        mov     address, sp                'Set "address" param or read_word to sp
                        call    #read_memory               'Read a byte from the stack into m_reg
                        or      pc, m_reg
                        rol     pc, #8
                        add     sp, #1                     'Bump up the stack pointer
                        and     sp, word_mask              'Ensure still 16 bits (roll over)
                        jmp     #fetch

jmp_conditional         call    #word_fetch                'Get instructions immediate word into "imm_word"
                        test    do_flag, #1 wz
              if_nz     mov     pc, address                'Set program counter to the immediate address
                        jmp     #fetch

'JMP - Jump
do_jmp                  call    #word_fetch
                        mov     pc, address                'Set program counter to the immediate address
                        jmp     #fetch

'CALL - Call
do_call                 call    #word_fetch
                        jmp     #call_conditional_tail

'RST - Restart
do_rst                  mov     scratch, m_reg
                        and     scratch, #%00111000
                        jmp     #call_scratch



'PUSH [BC|DE|HL] - Push Data On To Stack
do_push          '       mov     opc, m_reg
                        call    #get_reg_pair
                        sub     sp, #1                     'Decrement stack pointer first
                        mov     address,sp                 'Set "address" parameter of write_memory to stack pointer
                        mov     m_reg, tmp_h               'Set m_reg parameter of write_memory to "scratch"
                        call    #write_memory              'Write the high byte to the stack
                        sub     sp, #1                     'Decrement stack pointer again
                        mov     address,sp                 'Set "address" parameter of write_memory to stack pointer
                        mov     m_reg, tmp_l               'Set m_reg parameter of write_memory to "scratch" 
                        call    #write_memory              'Write the low byte to the stack
                        jmp     #fetch

'PUSH AF - Push Data On To Stack
do_push_af              sub     sp, #1                     'Decrement stack pointer first
                        mov     address,sp                 'Set "address" parameter of write_memory to stack pointer
                        mov     m_reg, a_reg               'Set m_reg parameter of write_memory to "scratch"
                        call    #write_memory              'Write the high byte to the stack
                        sub     sp, #1                     'Decrement stack pointer again
                        mov     address,sp                 'Set "address" parameter of write_memory to stack pointer
                        mov     m_reg, psw_reg             'Set m_reg parameter of write_memory to "scratch"
              '          and     m_reg, #%11010111         'Set and clear unused flags as per Intel manual
              '          or      m_reg, #%00000010 
                        call    #write_memory              'Write the low byte to the stack
                        jmp     #fetch

'POP [BC|DE|HL] - Pop Data From Stack
do_pop                  mov     opc, m_reg                 'Save the opcode
                        mov     address, sp                'Set "address" param or read_word to sp    
                        call    #read_memory               'Read a byte from the stack into m_reg
                        mov     tmp_l, m_reg
                        add     sp, #1                     'Bump up the stack pointer
                        and     sp, word_mask              'Ensure still 16 bits (roll over)
                        mov     address, sp                'Set "address" param or read_word to sp
                        call    #read_memory               'Read a byte from the stack into m_reg
                        mov     tmp_h, m_reg
                        add     sp, #1                     'Bump up the stack pointer
                        and     sp, word_mask              'Ensure still 16 bits (roll over)
                        mov     m_reg, opc
                        call    #put_reg_pair
                        jmp     #fetch

'POP AF - Pop Data From Stack
do_pop_af               mov     address, sp                'Set "address" param or read_word to sp    
                        call    #read_memory               'Read a byte from the stack into m_reg
               '         and     mem_reg, #%11010111       'Set and clear unused flags as per Intel manual
               '         or      mem_reg, #%00000010 
                        mov     psw_reg, m_reg
                        add     sp, #1                     'Bump up the stack pointer
                        and     sp, word_mask              'Ensure still 16 bits (roll over)
                        mov     address, sp                'Set "address" param or read_word to sp
                        call    #read_memory               'Read a byte from the stack into m_reg
                        mov     a_reg, m_reg
                        add     sp, #1                     'Bump up the stack pointer
                        and     sp, word_mask              'Ensure still 16 bits (roll over)
                        jmp     #fetch

'INX [BC|DE|HL] - Increment Register Pair
do_inx_rp               call    #get_reg_pair
                        add     tmp_l, #1
                        test    tmp_l, #$100 wz
              if_nz     add     tmp_h, #1
                        call    #put_reg_pair
                        jmp     #fetch

'DCX [BC|DE|HL] - Decrement Register Pair
do_dcx_rp               call    #get_reg_pair
                        sub     tmp_l, #1
                        test    tmp_l, #$100 wz
              if_nz     sub     tmp_h, #1
                        call    #put_reg_pair
                        jmp     #fetch

'DAD [BC|DE|HL] - Double Add Register Pair To HL                        
do_dad_rp               call    #get_reg_pair
                        add     l_reg, tmp_l
                        add     h_reg, tmp_h
                        test    l_reg, #256  wz
              if_nz     add     h_reg, #1
                        test    h_reg, #256 wz
                        muxnz   psw_reg, #carry_bit
                        and     h_reg, #$FF
                        and     l_reg, #$FF
                        jmp     #fetch

'XCHG - Exchange Registers
do_xchg                 xor     h_reg, d_reg              'Exchage HL with DE
                        xor     d_reg, h_reg
                        xor     h_reg, d_reg
                        xor     l_reg, e_reg
                        xor     e_reg, l_reg
                        xor     l_reg, e_reg
                        jmp     #fetch

'LXI HL - Load HL from Immediate
do_lxi_hl               call    #instruction_fetch        'Read immediate byte into 'm_reg'
                        mov     l_reg, m_reg 
                        call    #instruction_fetch        'Read immediate byte into 'm_reg'
                        mov     h_reg, m_reg
                        jmp     #fetch

'LXI DE - Load DE from Immediate
do_lxi_de               call    #instruction_fetch        'Read immediate byte into 'm_reg'
                        mov     e_reg, m_reg 
                        call    #instruction_fetch        'Read immediate byte into 'm_reg'
                        mov     d_reg, m_reg
                        jmp     #fetch

'Read the next word from pc and pc + 1 and place it in "address"
word_fetch              mov     address, pc
                        call    #read_memory               'Read low byte from (PC) into m_reg
                        add     pc,#1
                        mov     data, m_reg
                        mov     address, pc
                        call    #read_memory               'Read high byte from (PC) into m_reg
                        add     pc,#1
                        shl     m_reg, #8
                        or      data, m_reg
                        mov     address, data              'Return word in address
word_fetch_ret          ret

'Read an instruction from byte at pc into m_reg, increment pc by 1. 
instruction_fetch       mov     address, pc                'Set address param for read_memory
                        call    #read_memory               'Read the byte there 
                        add     pc, #1                     'Increment the program counter
                        and     pc, word_mask              'Ensure it is only 16 bits still (roll round)
instruction_fetch_ret   ret

'Read a byte from location given by HL into byte "data"
read_memory_hl          mov     address, h_reg             'Make an address out of h and l
                        shl     address, #8
                        add     address, l_reg             'Fall through  to read_memory (from address)
                        
'Read a byte from location given by "address" into byte m_reg
read_memory             cmp     address, ram_size wc,wz    'Is the given address within the RAM range ?
              if_b      jmp     #read_ram
                        call    #read_rom
                        jmp     #read_memory_ret
read_ram                mov     pointer, ram_base          'Point to the RAM
                        add     pointer, address           'Add address offset
                        rdbyte  m_reg, pointer             'Read the byte from there
read_memory_hl_ret
read_memory_ret         ret

read_rom                cmp     address, rom_start wc,wz   'Is the given address within the ROM range ?
              if_b      jmp     #read_rom_ret              'address below rom_base => return with "data" undefined
                        mov     rom_address, address
                        sub     rom_address, rom_start     'address becomes offset into ROM from start of ROM                  
                        mov     pointer, rom_base          'Point to the ROM
                        add     pointer, rom_address       'Add address offset                     
                        rdbyte  m_reg, pointer             'Read the byte from there
read_rom_ret            ret                                'Return

'Write the byte m_reg to location given by HL
write_memory_hl         mov     address, h_reg             'Make an address out of h and l
                        shl     address, #8
                        add     address, l_reg             'Fall through to write_memory (to address)
                        
'Write the byte m_reg to location given by "address"
write_memory            cmp     address, ram_size wc,wz    'Is the given address within the RAM range ?
              if_b      mov     pointer, ram_base          'Point to the RAM
              if_b      add     pointer, address           'Add address offset
              if_b      wrbyte  m_reg, pointer             'Write the byte there
write_memory_hl_ret
write_memory_ret        ret

'Load register pair (BC, DE, HL) specified by instruction into tmp_h and tmp_l
get_reg_pair            mov     rp_idx, m_reg
                        and     rp_idx, #%00_11_0000       'Isolate reg pair bits from instruction
                        shr     rp_idx, #3                 'Make into register file index
                        movs    get_th, rp_idx             'Modify get_th to get high register
                        add     rp_idx, #1
                        movs    get_tl, rp_idx             'Modify get_tl to get low register
get_th                  mov     tmp_h, 0-0                 'Get the regs
get_tl                  mov     tmp_l, 0-0
get_reg_pair_ret        ret

'Load tmp_h and tmp_l from register pair (BC, DE, HL) specified by instruction
put_reg_pair            mov     rp_idx, m_reg
                        and     rp_idx, #%00_11_0000       'Isolate reg pair bits from instruction
                        shr     rp_idx, #3                 'Make into register file index
                        movd    put_th, rp_idx             'Modify put_th to put high register
                        add     rp_idx, #1
                        movd    put_tl, rp_idx             'Modify put_tl to put low register
                        and     tmp_h, #$FF                'Keep things in range
                        and     tmp_l, #$FF                                     
put_th                  mov     0-0, tmp_h                 'Put the regs
put_tl                  mov     0-0, tmp_l
put_reg_pair_ret        ret

{{
' Display a byte on the leds for one second
display                 mov     dira, pins                 'Set Pins to output
                        shl     disp, #16                  'Get data into right bits
                        mov     outa, disp                 'Display it
        '               mov     time, cnt                  'Calculate delay time
        '               add     time, delay                 'Set 1 second delay here
        '               waitcnt time, 0                     'Wait
display_ret  ret


dump_regs
                        movd    dmp,#0
                        mov     reg_ptr, reg_base
                        mov     scratch, #12              
dmp                     wrlong  0-0,reg_ptr
                        add     reg_ptr,#4
                        add     dmp,dst_1
                        sub     scratch,#1
                        tjnz    scratch,#dmp                             
dump_regs_ret           ret

break                   call    #dump_regs
                        mov     scratch, #$FF
                        wrbyte  scratch, ios_op 
brk_not_rdy             rdbyte  scratch, ios_op wz
              if_nz     jmp     #brk_not_rdy
break_ret               ret         
}}


'Constant values
ram_size                long    (1024 * 24)
rom_start               long    $FF00                      'Altair ROM starts at this address
word_mask               long    $0000FFFF                  'Lower 16 bit mask                     
dst_1                   long    %1000000000                'LSB of destination field
pins                    long    $00800000                  'Propeller Demo Board LED pins

'Values to be setup by spin prior to launching COG
reg_base                long    0-0                        'HUB address of 11 longs into which registers may be dumped on break
ram_base                long    0-0                        'HUB address of the 8080 RAM area
rom_base                long    0-0                        'HUB address of the BOOT ROM image            
ios_port                long    0-0                        'Three longs for communicating with the I/O system
ios_op                  long    0-0                        '8080 port number, operation and data byte
ios_data                long    0-0
object_base             long    0-0                        'HUB address of this object 

'Working variables
scratch                 long    0
reg_ptr                 long    0
mem_ptr                 long    0
data                    long    0
address                 long    0
pointer                 long    0
rom_address             long    0
alu_reg                 long    0                        
opc                     long    0
dispatch_tab_addr       long    @dispatch_table
lmm_pc                  long    0
logic_op                long    0
aux                     long    0
operand                 long    0
rp_idx                  long    0
tmp_h                   long    0
tmp_l                   long    0
do_flag                 long    0


'Debugging feature variables
{{
disp                    long    0
single_step             long    0
time                    long    0
delay                   long    80_000_000 * 1
break_point             long    $FFFF
}}

' Translation table, 8080 alu op to Prop op
opTransTable            long    propADD                    'Prop ADD  for 8080 ADD and ADI
                        long    propADDX                   'Prop ADDX for 8080 ADC and ACI
                        long    propSUB                    'Prop SUB  for 8080 SUB and SUI
                        long    propSUBX                   'Prop SBB  for 8080 SUI and SBI
                        long    propAND                    'Prop AND  for 8080 ANA and ANI
                        long    propXOR                    'Prop XOR  for 8080 XRA and XRI
                        long    propOR                     'Prop OR   for 8080 ORA and ORI
                        long    propSUB                    'Prop SUB  for 8080 CMP and CPI

'1 Long free here!!

                        fit

'The start of code that is executed from HUB by the Large Memory Model Virtual Machine
'N.B. Can only jump or call to locations in COG from here. Don't try to add self modifying code.
'N.B. These instructions are reversed in order in memory prior to COG start and run by a reverse execution order VM
lmm_begin

'DAA - Decimal Adjust For Addition
do_daa                  mov     scratch, a_reg            'Isolate low nibble of accumulator
                        and     scratch, #$0F                        
                        cmp     scratch, #10 wc           'Test for greater than 9                
                        test    psw_reg, #aux_bit wz      'Test auxillary carry
            if_nz_or_nc add     a_reg, #$06               'Add 6 to accumulator if greater than 9 or AUX carry
                        add     scratch, #6               'Check for AUX carry from low nibble
                        test    scratch, #%00010000 wz
                        muxnz   psw_reg, #aux_bit         'and set AUX flag if so.
                        mov     scratch, a_reg            'Isolate high nibble of accumulator
                        shr     scratch, #4
                        cmp     scratch, #10 wc           'Test for greater than 9
                        test    psw_reg, #carry_bit wz    'Test carry flag
            if_nz_or_nc add     a_reg, #$60 wc            'Add (6 << 4) to accumulator if greater than 9 or AUX carry
                        test    a_reg, #$100 wz           'Check for 8 bit carry out
            if_nz       or      psw_reg, #carry_bit       'Set carry flag if so (N.B. Do NOT clear carry if not so)
                        and     a_reg, #$FF wz, wc
                        muxz    psw_reg, #zero_bit        'Set 8080 zero flag from props zero
                        muxnc   psw_reg, #parity_bit
                        test    a_reg, #128 wz
                        muxnz   psw_reg, #sign_bit
                        jmp     #fetch

'LXI BC - Load BC from Immediate
do_lxi_bc               call    #instruction_fetch        'Read immediate byte into 'm_reg'
                        mov     c_reg, m_reg 
                        call    #instruction_fetch        'Read immediate byte into 'm_reg'
                        mov     b_reg, m_reg
                        jmp     #fetch

'LXI SP - Load Stack Pointer from Immediate
do_lxi_sp               call    #instruction_fetch        'Read immediate byte into 'm_reg'
                        mov     sp, m_reg
                        ror     sp, #8
                        call    #instruction_fetch        'Read immediate byte into 'm_reg'
                        or      sp, m_reg
                        rol     sp, #8
                        jmp     #fetch

'DAD SP - Double Add Stack Pointer To HL                        
do_dad_sp               mov     scratch, sp
                        and     scratch, #$FF
                        add     l_reg, scratch
                        mov     scratch, sp
                        shr     scratch, #8
                        add     h_reg, scratch
                        test    l_reg, #256  wz
              if_nz     add     h_reg, #1
                        test    h_reg, #256 wz
                        muxnz   psw_reg, #carry_bit
                        and     h_reg, #$FF
                        and     l_reg, #$FF
                        jmp     #fetch

'XTHL - Exchange Stack
do_xthl                 mov     address, sp               'Set address param for read_memory to stack pointer
                        call    #read_memory              'Read byte from "address" into "m_reg"
                        xor     m_reg, l_reg               'Swap l reg and data
                        xor     l_reg, m_reg
                        xor     m_reg, l_reg
                        call    #write_memory             'Write m_reg to location "address" 
                        add     address, #1               'Set address param for read_memory to next stack location up
                        call    #read_memory              'Read byte from "address" into m_reg
                        xor     m_reg, h_reg               'Swap h reg and data
                        xor     h_reg, m_reg
                        xor     m_reg, h_reg
                        call    #write_memory             'Write the byte "data" to location "address" 
lmm_end                 jmp     #fetch

'Instruction dispatch table.
'This table contains the addresses of all the 8080 opcode handler functions in order of opcode.
'For handlers implemented in native assembler the COG address is used. (0 to 496)
'For handlers implemented in Large Memory Model (LMM)code the offset (@) from the start of this object is used (<496)
'adjusted for the fact that the LMM runs it's code in reverse order !
'The main dispatch loop will convert these to absolute HUB addreses for the LMM VM on the fly. 
'Notice that this table is used from HUB RAM as it will not fit in the cog. There is a small speed penalty
'in doing this (rdword) but not much compared to total instruction execution time.
dispatch_table          
{00000000}              word    do_nop  
{00000001}              word    (@lmm_end + @lmm_begin - @do_lxi_bc) 
{00000010}              word    do_stax_b  
{00000011}              word    do_inx_rp 
{00000100}              word    do_inr_b
{00000101}              word    do_dcr_b  
{00000110}              word    do_mvi_b  
{00000111}              word    do_rlc  
{00001000}              word    do_undocumented  
{00001001}              word    do_dad_rp  
{00001010}              word    do_ldax_b  
{00001011}              word    do_dcx_rp 
{00001100}              word    do_inr_c  
{00001101}              word    do_dcr_c  
{00001110}              word    do_mvi_c  
{00001111}              word    do_rrc  
{00010000}              word    do_undocumented  
{00010001}              word    do_lxi_de 
{00010010}              word    do_stax_d  
{00010011}              word    do_inx_rp 
{00010100}              word    do_inr_d  
{00010101}              word    do_dcr_d  
{00010110}              word    do_mvi_d  
{00010111}              word    do_ral  
{00011000}              word    do_undocumented  
{00011001}              word    do_dad_rp  
{00011010}              word    do_ldax_d  
{00011011}              word    do_dcx_rp 
{00011100}              word    do_inr_e  
{00011101}              word    do_dcr_e  
{00011110}              word    do_mvi_e  
{00011111}              word    do_rar  
{00100000}              word    do_undocumented  
{00100001}              word    do_lxi_hl 
{00100010}              word    do_shld  
{00100011}              word    do_inx_rp 
{00100100}              word    do_inr_h  
{00100101}              word    do_dcr_h  
{00100110}              word    do_mvi_h  
{00100111}              word    (@lmm_end + @lmm_begin - @do_daa)        
{00101000}              word    do_undocumented  
{00101001}              word    do_dad_rp  
{00101010}              word    do_lhld  
{00101011}              word    do_dcx_rp 
{00101100}              word    do_inr_l  
{00101101}              word    do_dcr_l  
{00101110}              word    do_mvi_l  
{00101111}              word    do_cma  
{00110000}              word    do_undocumented  
{00110001}              word    (@lmm_end + @lmm_begin - @do_lxi_sp) 
{00110010}              word    do_sta  
{00110011}              word    do_inx_sp  
{00110100}              word    do_inr_m  
{00110101}              word    do_dcr_m  
{00110110}              word    do_mvi_m  
{00110111}              word    do_stc  
{00111000}              word    do_undocumented  
{00111001}              word    (@lmm_end + @lmm_begin - @do_dad_sp)   
{00111010}              word    do_lda  
{00111011}              word    do_dcx_sp  
{00111100}              word    do_inr_a  
{00111101}              word    do_dcr_a  
{00111110}              word    do_mvi_a  
{00111111}              word    do_cmc  

'All the mov instructions
{01000000}              word    do_mov_b_b  
{01000001}              word    do_mov_b_c  
{01000010}              word    do_mov_b_d  
{01000011}              word    do_mov_b_e
{01000100}              word    do_mov_b_h
{01000101}              word    do_mov_b_l  
{01000110}              word    do_mov_b_m  
{01000111}              word    do_mov_b_a
{01001000}              word    do_mov_c_b  
{01001001}              word    do_mov_c_c  
{01001010}              word    do_mov_c_d 
{01001011}              word    do_mov_c_e  
{01001100}              word    do_mov_c_h  
{01001101}              word    do_mov_c_l  
{01001110}              word    do_mov_c_m  
{01001111}              word    do_mov_c_a
{01010000}              word    do_mov_d_b
{01010001}              word    do_mov_d_c  
{01010010}              word    do_mov_d_d  
{01010011}              word    do_mov_d_e  
{01010100}              word    do_mov_d_h  
{01010101}              word    do_mov_d_l  
{01010110}              word    do_mov_d_m  
{01010111}              word    do_mov_d_a
{01011000}              word    do_mov_e_b  
{01011001}              word    do_mov_e_c  
{01011010}              word    do_mov_e_d  
{01011011}              word    do_mov_e_e  
{01011100}              word    do_mov_e_h  
{01011101}              word    do_mov_e_l  
{01011110}              word    do_mov_e_m  
{01011111}              word    do_mov_e_a
{01100000}              word    do_mov_h_b  
{01100001}              word    do_mov_h_c  
{01100010}              word    do_mov_h_d  
{01100011}              word    do_mov_h_e  
{01100100}              word    do_mov_h_h  
{01100101}              word    do_mov_h_l  
{01100110}              word    do_mov_h_m  
{01100111}              word    do_mov_h_a
{01101000}              word    do_mov_l_b  
{01101001}              word    do_mov_l_c  
{01101010}              word    do_mov_l_d  
{01101011}              word    do_mov_l_e  
{01101100}              word    do_mov_l_h  
{01101101}              word    do_mov_l_l  
{01101110}              word    do_mov_l_m  
{01101111}              word    do_mov_l_a
{01110000}              word    do_mov_m_b  
{01110001}              word    do_mov_m_c  
{01110010}              word    do_mov_m_d  
{01110011}              word    do_mov_m_e  
{01110100}              word    do_mov_m_h  
{01110101}              word    do_mov_m_l  
{01110110}              word    do_halt  
{01110111}              word    do_mov_m_a
{01111000}              word    do_mov_a_b  
{01111001}              word    do_mov_a_c  
{01111010}              word    do_mov_a_d  
{01111011}              word    do_mov_a_e  
{01111100}              word    do_mov_a_h  
{01111101}              word    do_mov_a_l  
{01111110}              word    do_mov_a_m  
{01111111}              word    do_mov_a_a  

'All the reg/mem to accumulator instructions
{10000000}              word    do_add_b 
{10000001}              word    do_add_c
{10000010}              word    do_add_d      
{10000011}              word    do_add_e  
{10000100}              word    do_add_h  
{10000101}              word    do_add_l
{10000110}              word    do_add_m
{10000111}              word    do_add_a
{10001000}              word    do_adc_b
{10001001}              word    do_adc_c
{10001010}              word    do_adc_d
{10001011}              word    do_adc_e
{10001100}              word    do_adc_h
{10001101}              word    do_adc_l
{10001110}              word    do_adc_m
{10001111}              word    do_adc_a
{10010000}              word    do_sub_b
{10010001}              word    do_sub_c
{10010010}              word    do_sub_d
{10010011}              word    do_sub_e
{10010100}              word    do_sub_h
{10010101}              word    do_sub_l
{10010110}              word    do_sub_m
{10010111}              word    do_sub_a
{10011000}              word    do_sbb_b
{10011001}              word    do_sbb_c
{10011010}              word    do_sbb_d
{10011011}              word    do_sbb_e
{10011100}              word    do_sbb_h
{10011101}              word    do_sbb_l
{10011110}              word    do_sbb_m
{10011111}              word    do_sbb_a
{10100000}              word    do_ana_b
{10100001}              word    do_ana_c
{10100010}              word    do_ana_d
{10100011}              word    do_ana_e
{10100100}              word    do_ana_h
{10100101}              word    do_ana_l
{10100110}              word    do_ana_m
{10100111}              word    do_ana_a
{10101000}              word    do_xra_b
{10101001}              word    do_xra_c
{10101010}              word    do_xra_d
{10101011}              word    do_xra_e
{10101100}              word    do_xra_h
{10101101}              word    do_xra_l
{10101110}              word    do_xra_m
{10101111}              word    do_xra_a
{10110000}              word    do_ora_b
{10110001}              word    do_ora_c
{10110010}              word    do_ora_d
{10110011}              word    do_ora_e
{10110100}              word    do_ora_h
{10110101}              word    do_ora_l
{10110110}              word    do_ora_m
{10110111}              word    do_ora_a
{10111000}              word    do_cmp_b
{10111001}              word    do_cmp_c
{10111010}              word    do_cmp_d
{10111011}              word    do_cmp_e
{10111100}              word    do_cmp_h
{10111101}              word    do_cmp_l
{10111110}              word    do_cmp_m
{10111111}              word    do_cmp_a

{11000000}              word    do_rnz  
{11000001}              word    do_pop
{11000010}              word    do_jnz  
{11000011}              word    do_jmp 
{11000100}              word    do_cnz 
{11000101}              word    do_push
{11000110}              word    do_adi  
{11000111}              word                    do_rst  
{11001000}              word    do_rz  
{11001001}              word    do_ret  
{11001010}              word    do_jz 
{11001011}              word    do_undocumented  
{11001100}              word    do_cz  
{11001101}              word    do_call   
{11001110}              word    do_aci  
{11001111}              word                    do_rst  
{11010000}              word    do_rnc  
{11010001}              word    do_pop
{11010010}              word    do_jnc  
{11010011}              word    do_out  
{11010100}              word    do_cnc  
{11010101}              word    do_push
{11010110}              word    do_sui  
{11010111}              word                    do_rst  
{11011000}              word    do_rc  
{11011001}              word    do_undocumented  
{11011010}              word    do_jc  
{11011011}              word    do_in  
{11011100}              word    do_cc  
{11011101}              word    do_undocumented  
{11011110}              word    do_sbi  
{11011111}              word                    do_rst  
{11100000}              word    do_rpo  
{11100001}              word    do_pop
{11100010}              word    do_jpo  
{11100011}              word    (@lmm_end + @lmm_begin - @do_xthl)  
{11100100}              word    do_cpo  
{11100101}              word    do_push
{11100110}              word    do_ani  
{11100111}              word                    do_rst  
{11101000}              word    do_rpe  
{11101001}              word    do_pchl  
{11101010}              word    do_jpe  
{11101011}              word    do_xchg   
{11101100}              word    do_cpe 
{11101101}              word    do_undocumented  
{11101110}              word    do_xri  
{11101111}              word                    do_rst  
{11110000}              word    do_rp  
{11110001}              word    do_pop_af   
{11110010}              word    do_jp 
{11110011}              word    do_di  
{11110100}              word    do_cp  
{11110101}              word    do_push_af   
{11110110}              word    do_ori  
{11110111}              word                    do_rst  
{11111000}              word    do_rm  
{11111001}              word    do_sphl  
{11111010}              word    do_jm  
{11111011}              word    do_ei  
{11111100}              word    do_cm  
{11111101}              word    do_undocumented  
{11111110}              word    do_cpi  
{11111111}              word                    do_rst

'The end  

          