''*************************************
''*  Altair Simulator Memory v2.3     *
''*  (C) 2009 Michael Rychlik         *
''*************************************

' v1.0 - 01 Oct 2007 - original version
' v1.1 - 09 Dec 2007 - No change just bump verion inline with other modules
' v1.2 - 09 Mar 2008 - No change, just bump version 
' v1.3 - 20 Mar 2008 - Added Altair ROM area loaded from file BOOTDSK.COM
'                      ROM is accessed by CBIOSX to get the boot up disk number
'                      otherwise is not usd as we have all of CP/M loaded already
'                    - Added the instructions set test program (T8080.com)
'                      to the Transien Program Area (TPA). 
' v1.4 - 21 Mar 2008 - Bumped version.
' v2.0 - 22 Mar 2008 - Removed loading of CCP, BDOS and CBIOSX into DAT section.
'                    - Moved ram_8080 to VAR section.
'                    - With this change the CPU runs through RAM from 0000 at boot, executing NOPS 00
'                      until it hits the DSKBOOT bootloader ROM. So bingo! it boots from SD card.
'                    - The test T8080.COM is no longer loaded or run on start up. It is no longer
'                      required to even pad out RAM upto the CCP.
' v2.1 - 21 Apr 2008 - Moved ram_8080 back to DAT section so that we can run progs embedded in file statements here.
'                    - Provided some options (by commenting in/out) for running CP/M from files included here,
'                    - running the 8080 instruction test progrsm prior to CP/M or booting CP/M from SD card.
' v2.2 - 14 Jan 2009 - Added option to run MITS Altair Extended Basic see OPTIONS below.
' v2.3 - 16 Jan 2009 - Set to run 4KBASIC
                                      
CON

'The following taken from the SIMH Altair simulator cfgccp.lib
'WARNING !!! Change msize is altair_mem also if you change it here !!!


 
msize =    24          ' adjust per installed system memory (K bytes)
bioslen = $d00         ' adjust as bios changes are made (if necessary)
ccplen = $800          ' cp/m constant
bdoslen = $e00         ' cp/m constant

' cp/m image size (rounded up to next 1k boundary)
cpmlen = (ccplen + bdoslen + bioslen + $3ff) / $400

ccpph   =     (msize - cpmlen) * 1024         ' ccp start address
bdosph  =     ccpph + ccplen                  ' bdos start address
biosph  =     ccpph + ccplen + bdoslen        ' bios start address

bdos    = bdosph + 6                          ' Entry point for BDOS



VAR                   
  'The RAM used by the CPU emulator
'  byte ram_8080[msize * 1024]

  
PUB init : Success


PUB get_ram_base : ram_base
  ram_base := @ram_8080

PUB get_rom_base : rom_base
  rom_base := @rom_8080

PUB get_size : size
  size := msize * 1024


                                 
DAT

' OPTIONS: Use the following definition of RAM_8080 for running MITS Altair basic or the one below for CP/M
'
RAM_8080
              file "minol22.bin"        ' Use "4kbas.bin", "8kbas.bin" or "exbas.bin"
              BYTE 0[20 * 1024]       ' Adjust this down for bigger basics to fit HUB RAM

ram_8080_end

{
ram_8080      'OPTIONS: uncomment one (and only one) of the three lines (BYTE ....) below as desired: 

              'Un-comment line below to start CPM directly from RAM image, no disks (SD card) required for boot up
              'BYTE $C3, biosph & $FF, (biosph >> 8) & $FF

              'Un-comment line below to run the T8080 instruction set test first, no disks required
              BYTE $C3, tpa & $FF, (tpa >> 8) & $FF

              'Un-comment line below to run the disk boot code from ROM space and boot CP/M from disk (SD card)
              'BYTE $C3, $00, $FF
              
              
              'Place a jump to BDOS at location $0005
              BYTE 0[2]
              BYTE $C3, bdos & $FF, (bdos >> 8) & $FF
              BYTE 0[8]


              'Pad out with zeros until the start of the Transient Program Area at $100
              BYTE 0[15 * 16]

              'The Transient Program Area (TPA)
              'Fill this with T8080 an instruction set test prog. that is assembled to be
              'the correct size to fill RAM up to the CPP address.
tpa           file "T8080.COM"   'file "EX.COM" ' 
              
              'Console Command Processor compiled from original Digital Research source.
              file "CCP.COM"

              'Basic Disk Operating System compiled from original Digital Resarch source
              file "BDOS.COM"

              'Customized BIOS of the SIMH Altair Emulator Project
              file "CBIOSX.COM"
ram_8080_end

}
              'The Altair boot rom. This read as rom located $FF00 by the cpu emulator
              'This file is the result of assembling DSKBOOT.MAC inside the SIMH Altair 8800 simulator.
rom_8080      file "DSKBOOT.COM"


        